// libraries
require('colors');
var Osc = require('node-osc');
var config = require('./config');
var api = require('./src/api')(config);
var _ = require('lodash');

/*********************************************************************
    Initialize
**********************************************************************/
// Create Client to send messages
var client = new Osc.Client(config.ip, config.port.send);
console.log('Sending OSC messages to '+config.ip.cyan+":"+(config.port.send+"").cyan);

// Create Server to receive messages
var server = new Osc.Server(config.port.receive,'0.0.0.0');
console.log('Receving OSC messages from '+(config.port.receive+"").cyan);

// For info, print optimizations
console.log('Optimizations:',JSON.stringify(config.optimize,null,4).cyan);

/*********************************************************************
                  ON CALLBACK (Firebase), SEND OSC
**********************************************************************/
api.callback(function(){
	var args = _.toArray(arguments);
	args = _.map(args,function(val){ return _.isUndefined(val) || _.isNull(val)? -1: val; });
	client.send.apply(client,args);

	if(!config.optimize.disable_log) {
		var cmd = "rc > osc".green;
		if(args[0] ) args[0] = ("/"+args[0]).cyan;
		args.unshift(cmd);
		console.log.apply(console,args);
	}
});

/********************************************************************
					ON OSC -> CALL API (update Firebase)
*********************************************************************/
server.on("message", function (msg, rinfo) {
	if(msg[0] && msg[0] == '#bundle') {
		msg = msg[2];
	}
	if((msg+"").toLowerCase() == "bang") {
		return;
	}
	if(msg[0]+"" == "/log"){
		config.optimize.disable_log = msg[1] !== "true" && msg[1] !== true;
		console.log("Log is "+(config.optimize.disable_log?'disabled':'enabled'));
		return;
	}
	var method = msg[0]? msg[0].substr(1): "command is missing";
	if(api[method]) {
		api[method].apply(api,msg.splice(1));
	} else {
		console.log('ERROR   received unknown command from OSC '.yellow+(msg[0]+"").red);
	}

});

