var nodejs = typeof window === 'undefined';
var Firebase = nodejs? require('firebase'): window.Firebase;
var _ = require('lodash');

var CONFIG_DEFAULTS = {
  "ip":"127.0.0.1",
  "url":"https://tin-men.firebaseio.com/1-2/",
	"port": {
		"send": 6666,
		"receive": 6667
	},
  "optimize": {
    "disable_log": true,
    "dont_send_camelstats": false,
    "max_messages_per_second": {
      "camel":200,
      "throw":100,
    }
  }
};

module.exports = function(config) {
	_.defaults(config,CONFIG_DEFAULTS);
	var optimize = config.optimize || {};
	var url = config.url;
  	var ref = new Firebase(url);
  	console.log('Firebase version [osc]:',url);

	/**
	Calculation - enrich incoming data.
	**/
	var calculate = {
		count: function(i){
			var values = {};
			_.each(DATA,function(data,uuid){
				if(data.button && data.button[i]) {
					var value = data.button[i];
					if(!values[value]) {
						values[value] = 0;
					}
					values[value]++;
				}
			});
			var args = ['count','interface',i];
			_.each(values,function(count,value){
				args.push(value);
				args.push(count);
			});
			emit.apply(emit,args);
		},
		camelstats: function(){
			var aN = 0,aVel = 0,aAvg = 0,bN = 0,bVel = 0,bAvg = 0;
			_.each(DATA,function(data,uuid){
				if(data.camel) {
					if(data.camel.who == "a") {
						aN++;
						aVel += data.camel.velocity || 0;
					} else if(data.camel.who == "b"){
						bN++;
						bVel += data.camel.velocity || 0;
					}
				}
			});
			if(aN > 0) aAvg = aVel / aN;
			if(bN > 0) bAvg = bVel / bN;
			emit("camelstats",aVel,aN,aAvg,bVel,bN,bAvg);
		}
	};

	/**
	Show State (Handle Show State Changes)
	**/
	var interfaceRef = ref.child("interface");
	interfaceRef.on('value',function(snapshot){
		emit('interface',snapshot.val());
	});

	var messagesRef = ref.child("messages").limitToLast(10);
	messagesRef.on('child_added',function(snapshot){
		var data = snapshot.val();
		emit("message",data.message,data.name,data.uuid,data.time);
	});

	var lightmasterRef = ref.child("lightmaster");
	lightmasterRef.on('value',function(snapshot){
		emit('lightmaster',snapshot.val());
	});

	var lightsRef = ref.child("lights");
	lightsRef.on('value',function(snapshot){
		args = snapshot.val() || [0,0,0];
		args.unshift('lights');
		emit.apply(emit,args);
	});

	/**
	User State (Handle User Property Change)
	**/
	var userHandler = {
		button: function(value,name,uuid) {
			_.each(value,function(val,i){
				if(!DATA[uuid].button || val !== DATA[uuid].button[i]) {
					emit("button",val,name,uuid);
					if(!DATA[uuid].button)
						DATA[uuid].button = {};
					DATA[uuid].button[i] = val;
					calculate.count(i);
				}
			});
		},
		contest: function(value,name,uuid) {
			emit("contest",value[0] || 0,value[1] || 0,value[2] || 0,name,uuid);
		},
		camel: function(value,name,uuid) {
			emit("camel",value.who,value.velocity || 0,name,uuid);
			DATA[uuid].camel = value;
			if(!optimize.dont_send_camelstats){
				calculate.camelstats();
			}
		},
		'throw': function(value,name,uuid) {
			emit("throw",value.what || "",value.velocity || 0,name,uuid);
		}
	};

	/***
	API
	***/
	var DATA = {};
	var API = {
		users: DATA, // reference to all User data
		callback: function(_emit) {
			emit = _emit;
		},
		reset: function(){
			_.each(DATA,function(data,uuid){
				ref.child("users/"+uuid+"/button").remove();
			});
		},
		// TODO - update for new APP!
		lightmaster: function(uuid) {
			lightmasterRef.set(uuid);
		},
		// TODO - already works for new APP, but 'currentRoute' is preffered!
		'interface': function(id) {
			interfaceRef.set(1.0*id);
		},
		chat: function(msg,name,uuid) {
		    if(msg && msg.trim().length > 0) {
		      var data = {
		        time: (new Date()).getTime(),
		        uuid: uuid || 0,
		        name: name || "anonymous",
		        message: msg
		      };
		      ref.child("messages").push(data);
		      //ref.child("history").push(data);
		    }
		},
		getusercount: function() {
			var count = 1.0 * _.size(DATA);
			emit("usercount",count);
		},
		getusers: function() {
			var names = _.pluck(DATA,'name');
			var uuids = _.keys(DATA);
			var args = _.flatten(_.zip(uuids,names));
			args.unshift("users");
			emit.apply(emit,args);
		},
		sample: function(sample,sync,uuidOrVoices) {
			if(!uuidOrVoices && sync) uuidOrVoices = sync;
			if(!sample || !uuidOrVoices) {
				console.log("error: no sample, uuid or # voices!");
				return;
			}
			var uuids;
			if(/([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}|[a-z0-9]{12})/.test(uuidOrVoices)){
				uuids = [uuidOrVoices];
			} else {
				uuids = _.shuffle(_.keys(DATA));
				uuids = uuids.slice(0,uuidOrVoices);
			}

			console.log('sample',sample,uuids);
			_.each(uuids,function(uuid){
				ref.child("users/"+uuid+"/sound").set(sample);
			});
		},
		clearchat: function() {
			ref.child("messages").remove();
		},
		// New app completely reinstalls all files.
		refresh: function() {
			var value = (new Date()).getTime()*1.0 + Math.round(Math.random() * 100000);
			ref.child("refresh").set(value);
		},
		// TODO - update for new APP!
		buttons: function() {
			ref.child("buttons").set(_.toArray(arguments));
			var values = _.toArray(arguments).map(function(text){
				return {"content":text};
			});
			console.log(values);
		},
		nuke: function() {
			API.refresh();
			ref.child("users").remove();
		},
		// TODO - update for new APP!
		setcontest: function(i) {
			if(i >= 0 && i <= 2) {
				i = i * 1.0;
				var files = _.toArray(arguments).slice(1);
				ref.child("contest/"+i).set(files);
			}
		},
		simulate: function(f,secs,cmd) {
			var args = _.toArray(arguments).slice(2);
			var interval = 1/f;
			var n = secs*f;
			var timer;
			timer = setInterval(function(){
				emit.apply(emit,args);
				n--;
				if(n === 0){
					clearInterval(timer);
					console.log('simulation ended '+args[0]);
				}
			},interval);
		},
		stresstest: function() {
			for(var i = 0; i < 1000; i++){
				var uuid = 'test'+i;
				DATA[uuid] = {
					name: uuid,
					camel: {
						who: Math.random() > 0.5? 'a':'b',
						velocity: Math.ceil(Math.random() * 127)
					}
				};
			}
			var n = 1000;
			console.log('stresstest starting');
			var timer = setInterval(function(){
				if(n > 0) {
					for(var i = 0; i < 1000; i++){
						DATA['test'+i].camel.velocity = Math.ceil(Math.random() * 127);
						update('camel',DATA[uuid].camel,uuid,uuid);
					}
				}
				calculate.camelstats();
				calculate.count();
				if(n === 0) {
					clearInterval(timer);
					console.log('stresstest done');
					for(var i = 0; i < 1000; i++) {
						delete DATA['test'+i];
					}
				}
				n--;
			},10);
		}
	};

	/**********************************************************************
	*                 OLD / NEW API HACKS
	***********************************************************************/
	// The new API does not REMOVE the entire user state anymore.
	// Instead, it keeps track of user online status in a seperate hashmap.
	var onlineRef = ref.child('online');
	onlineRef.on('child_changed',function(snap){
		var uuid = snap.key();
		var name = DATA[uuid] && DATA[uuid].name? DATA[uuid].name: '';
		var isOnline = !!snap.val();
		if(isOnline) {
			online(name,uuid);
		} else {
			offline(name,uuid);
		}
	});

	/**********************************************************************
	*                 IMPLEMENTATION INTERNALS
	***********************************************************************/
	/**
	Event Emitter (i.e. emits updates from Firebase)
	**/
	var emit = function(){
		var args = _.toArray(arguments);
		var cmd = "rc > osc";
		if(args[0]) args[0] = "/"+args[0];
		args.unshift(cmd);
		console.log.apply(console,args);
	};

	/**
	Handle online, updates, offline
	**/
	function online(name,uuid) {
		if(!DATA[uuid]) DATA[uuid] = {};
		DATA[uuid].online = true;
		emit("online",name,uuid);
	}
	function offline(name,uuid) {
		if(!DATA[uuid]) DATA[uuid] = {};
		DATA[uuid].online = false;
		emit("offline",name,uuid);
	}
	function update(key,value,name,uuid) {
		if(userHandler[key]) {
			userHandler[key](value,name,uuid);
		}
	}

	/**
	Monitor change
	**/
	var dataRef = ref.child("users");
	dataRef.on('child_added',function(snapshot){
		var uuid = snapshot.key();
		var data = snapshot.val() || {};
		var name = data.name || "anonymous";
		if(!DATA[uuid]) DATA[uuid] = {};
		online(name,uuid);
		_.each(data,function(val,key){
			update(key,val,name,uuid);
		});
		DATA[uuid] = data;
	});
	dataRef.on('child_changed',function(snapshot){
		var data = snapshot.val();
		var uuid = snapshot.key();
		var name = DATA[uuid].name;
		_.each(data,function(val,key){
			if(!_.isEqual(DATA[uuid][key],val)){
				update(key,val,name,uuid);
				DATA[uuid][key] = val;
			}
		});
	});
	dataRef.on('child_removed',function(snapshot){
		var uuid = snapshot.key();
		var name = DATA[uuid].name;
	});

	// throttle optimizations
	calculate.camelstats = _.throttle(calculate.camelstats,2500,{leading:false,trailing:true});
	calculate.count = _.throttle(calculate.count,1000,{leading:false,trailing:true});
	if(optimize.max_messages_per_second.camel){
		userHandler.camel = _.throttle(userHandler.camel,1000 / optimize.max_messages_per_second.camel,{leading:true,trailing:true});
	}
	if(optimize.max_messages_per_second['throw']){
		userHandler['throw'] = _.throttle(userHandler['throw'],1000 / optimize.max_messages_per_second['throw'],{leading:true,trailing:true});
	}

	return API;
};